from abc import abstractmethod
from typing import TypeVar, Generic
from collections.abc import Sequence, Collection

import numpy as np
import scipy.optimize as spo

from metametric.core.constraint import MatchingConstraint

T = TypeVar("T")


class MatchingProblem(Generic[T]):
    """A matching between two collections of objects."""

    def __init__(self, x: Sequence[T], y: Sequence[T], gram_matrix: np.ndarray):
        self.x = x
        self.y = y
        self.gram_matrix = gram_matrix

    @abstractmethod
    def solve(self) -> tuple[float, Collection[tuple[int, int, float]]]:
        """Solves the matching problem."""
        raise NotImplementedError


class AssignmentProblem(MatchingProblem[T]):
    def __init__(self, x: Sequence[T], y: Sequence[T], gram_matrix: np.ndarray, constraint: MatchingConstraint):
        super().__init__(x, y, gram_matrix)
        self.constraint = constraint

    def solve(self) -> tuple[float, Collection[tuple[int, int, float]]]:
        m = self.gram_matrix
        if self.constraint == MatchingConstraint.ONE_TO_ONE:
            row_idx, col_idx = spo.linear_sum_assignment(
                cost_matrix=m,
                maximize=True,
            )
            total = m[row_idx, col_idx].sum()
            matching = [(i.item(), j.item(), m[i, j].item()) for i, j in zip(row_idx, col_idx)]
            return total, matching
        if self.constraint == MatchingConstraint.ONE_TO_MANY:
            total = m.max(axis=0).sum().item()
            selected_x = m.argmax(axis=0)
            matching = [(selected_x[j].item(), j, m[selected_x[j], j].item()) for j in range(m.shape[1])]
            return total, matching
        if self.constraint == MatchingConstraint.MANY_TO_ONE:
            total = m.max(axis=1).sum().item()
            selected_y = m.argmax(axis=1)
            matching = [(i, selected_y[i].item(), m[i, selected_y[i]].item()) for i in range(m.shape[0])]
            return total, matching
        if self.constraint == MatchingConstraint.MANY_TO_MANY:
            total = m.sum().item()
            matching = [(i, j, m[i, j].item()) for i in range(m.shape[0]) for j in range(m.shape[1])]
            return total, matching
        raise ValueError(f"Invalid constraint: {self.constraint}")



"""Metric suites are collections of metrics that are computed together."""

from dataclasses import dataclass
from typing import Callable, Protocol, TypeVar, Optional, Any
from collections.abc import Sequence

from metametric.core.matching import Hook
from metametric.core.metric import Metric
from metametric.core.reduction import Reduction
from metametric.core.state import MetricState, MultipleMetricStates, SingleMetricState

T = TypeVar("T", contravariant=True)


class Aggregator(Protocol[T]):
    """Metric aggregator for computing metrics on a stream of predictions and references."""

    @property
    def state(self) -> MetricState[T]:
        """The internal state of the aggregator."""
        raise NotImplementedError()

    @property
    def hooks(self) -> Optional[dict[str, Hook[Any]]]:
        """Get the hooks for the aggregator."""
        raise NotImplementedError

    def update_single(self, pred: T, ref: T) -> None:
        """Update the aggregator with a single prediction and its reference."""
        self.state.update_single(pred, ref, self.hooks)

    def update_batch(self, preds: Sequence[T], refs: Sequence[T]) -> None:
        """Update the aggregator with a batch of predictions and their references."""
        self.state.update_batch(preds, refs, self.hooks)

    def reset(self) -> None:
        """Reset the aggregator to its initialization state."""
        self.state.reset()

    def compute(self) -> dict[str, float]:
        """Compute the metrics from the aggregator."""
        raise NotImplementedError()


class MetricSuite(Protocol[T]):
    """Metric suites are collections of metrics that are computed together."""

    def new(self, hooks: Optional[dict[str, Hook[Any]]] = None) -> Aggregator[T]:
        """Create a new aggregator for the metric suite."""
        raise NotImplementedError()

    def with_extra(self, extra: Callable[[dict[str, float]], dict[str, float]]) -> "MetricSuite[T]":
        """Add extra metrics to the metric suite."""
        return MetricSuiteWithExtra(self, extra)


@dataclass
class MetricFamily(MetricSuite[T]):
    """A collection of metrics that shared the same internal state.

    For example, the precision, recall, and F-1 of event detection should be computed together within one family.
    """

    def __init__(self, metric: Metric[T], reduction: Reduction):
        self.metric = metric
        self.reduction = reduction

    def new(self, hooks: Optional[dict[str, Hook[Any]]] = None) -> Aggregator[T]:
        return MetricFamilyAggregator(self, hooks)


class MetricFamilyAggregator(Aggregator[T]):
    """Aggregator for a metric family."""

    def __init__(self, family: MetricFamily[T], hooks: Optional[dict[str, Hook[Any]]] = None):
        self.family = family
        self._state = SingleMetricState(family.metric)
        self._hooks = hooks

    @property
    def state(self) -> MetricState[T]:
        return self._state

    @property
    def hooks(self) -> Optional[dict[str, Hook[Any]]]:
        return self._hooks

    def compute(self) -> dict[str, float]:
        return self.family.reduction.compute(self._state)


class MultipleMetricFamilies(MetricSuite[T]):
    """A collection of metric families, whose internal states are separate."""

    def __init__(self, families: dict[str, MetricSuite[T]]):
        self.families = families

    def new(self, hooks: Optional[dict[str, Hook[Any]]] = None) -> Aggregator[T]:
        return MultipleMetricFamiliesAggregator(self, hooks)


class MultipleMetricFamiliesAggregator(Aggregator[T]):
    """Aggregator for multiple metric families."""

    def __init__(self, coll: MultipleMetricFamilies[T], hooks: Optional[dict[str, Hook[Any]]] = None):
        self.aggs: dict[str, Aggregator[T]] = {name: family.new() for name, family in coll.families.items()}
        self._hooks = hooks

    @property
    def state(self) -> MetricState[T]:
        return MultipleMetricStates({name: agg.state for name, agg in self.aggs.items()})

    @property
    def hooks(self) -> Optional[dict[str, Hook[Any]]]:
        return self._hooks

    def compute(self) -> dict[str, float]:
        metrics = {
            f"{name}-{key}" if key != "" else name: value
            for name, agg in self.aggs.items()
            for key, value in agg.compute().items()
        }
        return metrics


class MetricSuiteWithExtra(MetricSuite[T]):
    """A metric suite with extra metrics."""

    def __init__(self, original: MetricSuite[T], extra: Callable[[dict[str, float]], dict[str, float]]):
        self.original = original
        self.extra = extra

    def new(self, hooks: Optional[dict[str, Hook[Any]]] = None) -> Aggregator[T]:
        return WithExtraAggregator(self, hooks)


class WithExtraAggregator(Aggregator[T]):
    """Aggregator for a metric suite with extra metrics."""

    def __init__(self, coll: MetricSuiteWithExtra[T], hooks: Optional[dict[str, Hook[Any]]] = None):
        self.agg = coll.original.new(hooks)
        self.extra = coll.extra
        self._hooks = hooks

    @property
    def state(self) -> MetricState[T]:
        return self.agg.state

    @property
    def hooks(self) -> Optional[dict[str, Hook[Any]]]:
        return self._hooks

    def compute(self) -> dict[str, float]:
        metrics = self.agg.compute()
        metrics.update(self.extra(metrics))
        return metrics



"""Metric derivation with matching constraints."""

from dataclasses import is_dataclass
from typing import TypeVar, Union
from collections.abc import Collection, Sequence
from collections import defaultdict

import numpy as np

from metametric.core._ilp import ILPMatchingProblem
from metametric.core.constraint import MatchingConstraint
from metametric.core.graph import Graph, _reachability_matrix
from metametric.core.matching import Match, Matching, Path
from metametric.core.metric import DiscreteMetric, Metric
from metametric.core._problem import AssignmentProblem


C = TypeVar("C")
T = TypeVar("T")


def _matching_from_triples(
    original_x: C,
    original_y: C,
    score: float,
    x: Sequence[T],
    y: Sequence[T],
    matches: Collection[tuple[int, int, float]],
) -> Matching:
    def _matching():
        yield Match(Path(), original_x, Path(), original_y, score)
        for i, j, s in matches:
            yield Match(Path().prepend(i), x[i], Path().prepend(j), y[j], s)

    return Matching(_matching())


class SetMatchingMetric(Metric[Collection[T]]):
    """A metric derived from the matching of two sets."""

    def __init__(self, inner: Metric[T], constraint: Union[str, MatchingConstraint] = MatchingConstraint.ONE_TO_ONE):
        self.inner = inner
        self.constraint = MatchingConstraint.from_str(constraint) if isinstance(constraint, str) else constraint

    def compute(self, x: Collection[T], y: Collection[T]) -> tuple[float, Matching]:
        """Score two sets of objects."""
        original_x, original_y = x, y
        x, y = list(x), list(y)
        x_is_empty = len(x) == 0
        y_is_empty = len(y) == 0
        if x_is_empty and y_is_empty:
            return 1.0, Matching([Match(Path(), x, Path(), y, 1.0)])
        elif x_is_empty or y_is_empty:
            return 0.0, Matching([])
        elif isinstance(self.inner, DiscreteMetric) and self.constraint == MatchingConstraint.ONE_TO_ONE:
            intersection = set(x) & set(y)
            score = len(intersection)

            def _matching():
                yield Match(Path(), x, Path(), y, score)
                x_indices, y_indices = defaultdict(list), defaultdict(list)
                for i, u in enumerate(x):
                    x_indices[u].append(i)
                for j, v in enumerate(y):
                    y_indices[v].append(j)
                for k in intersection:
                    for i, j in zip(x_indices[k], y_indices[k]):
                        yield Match(Path().prepend(i), x[i], Path().prepend(j), y[j], 1.0)

            return score, Matching(_matching())
        else:
            m = self.inner.gram_matrix(x, y)
            score, triples = AssignmentProblem(x, y, m, self.constraint).solve()
            return score, _matching_from_triples(original_x, original_y, score, x, y, triples)

    def score_self(self, x: Collection[T]) -> float:
        """Score a set of objects with itself."""
        x = list(x)
        if len(x) == 0:
            return 1.0
        elif self.constraint == MatchingConstraint.MANY_TO_MANY:
            return self.inner.gram_matrix(x, x).sum()
        elif self.constraint == MatchingConstraint.ONE_TO_ONE:
            return sum(self.inner.score_self(u) for u in x)
        else:
            return self.score(x, x)


class SequenceMatchingMetric(Metric[Sequence[T]]):
    """A metric derived from the matching of two sequences."""

    def __init__(self, inner: Metric[T], constraint: Union[str, MatchingConstraint] = MatchingConstraint.ONE_TO_ONE):
        self.inner = inner
        self.constraint = MatchingConstraint.from_str(constraint) if isinstance(constraint, str) else constraint

    def compute(self, x: Sequence[T], y: Sequence[T]) -> tuple[float, Matching]:
        m = self.inner.gram_matrix(x, y)
        f = np.zeros([m.shape[0] + 1, m.shape[1] + 1])
        for i in range(m.shape[0] + 1):
            for j in range(m.shape[1] + 1):
                if i == 0 or j == 0:
                    f[i, j] = 0
                else:
                    f[i, j] = max(f[i - 1, j - 1] + m[i - 1, j - 1], f[i - 1, j], f[i, j - 1])
                    if self.constraint == MatchingConstraint.ONE_TO_MANY:
                        f[i, j] = max(f[i, j], f[i, j - 1] + m[i - 1, j - 1])
                    elif self.constraint == MatchingConstraint.MANY_TO_ONE:
                        f[i, j] = max(f[i, j], f[i - 1, j] + m[i - 1, j - 1])
                    elif self.constraint == MatchingConstraint.MANY_TO_MANY:
                        f[i, j] = max(f[i, j], f[i, j - 1] + m[i - 1, j - 1], f[i - 1, j] + m[i - 1, j - 1])
        return f[-1, -1].item(), Matching([])  # TODO: implement matching

    def score_self(self, x: Sequence[T]) -> float:
        if self.constraint == MatchingConstraint.ONE_TO_ONE:
            return sum(self.inner.score_self(u) for u in x)
        else:
            return self.score(x, x)


class GraphMatchingMetric(Metric[Graph[T]]):
    """A metric derived from the matching of two graphs (including trees, DAGs, and general graphs)."""

    def __init__(self, inner: Metric[T], constraint: Union[str, MatchingConstraint] = MatchingConstraint.ONE_TO_ONE):
        self.inner = inner
        self.constraint = MatchingConstraint.from_str(constraint) if isinstance(constraint, str) else constraint

    def compute(self, x: Graph[T], y: Graph[T]) -> tuple[float, Matching]:
        x_nodes = list(x.nodes())
        y_nodes = list(y.nodes())
        gram_matrix = self.inner.gram_matrix(x_nodes, y_nodes)
        x_reach = _reachability_matrix(x)
        y_reach = _reachability_matrix(y)

        problem = ILPMatchingProblem(x_nodes, y_nodes, gram_matrix, has_vars=False)
        problem.add_matching_constraint(self.constraint)
        problem.add_monotonicity_constraint(x_reach, y_reach)
        score, matches = problem.solve()
        return score, _matching_from_triples(x, y, score, x_nodes, y_nodes, matches)


class LatentSetMatchingMetric(Metric[Collection[T]]):
    """A metric derived to support matching latent variables defined in structures."""

    def __init__(
        self,
        cls: type[T],
        inner: Metric[T],
        constraint: Union[str, MatchingConstraint] = MatchingConstraint.ONE_TO_ONE,
    ):
        if is_dataclass(cls):
            self.cls = cls
        else:
            raise ValueError(f"{cls} has to be a dataclass.")
        self.inner = inner
        self.constraint = MatchingConstraint.from_str(constraint) if isinstance(constraint, str) else constraint

    def compute(self, x: Collection[T], y: Collection[T]) -> tuple[float, Matching]:
        """Score two collections of objects."""
        original_x, original_y = x, y
        x = list(x)
        y = list(y)

        x_is_empty = len(x) == 0
        y_is_empty = len(y) == 0

        if x_is_empty and y_is_empty:
            return 1.0, Matching([Match(Path(), x, Path(), y, 1.0)])
        elif x_is_empty or y_is_empty:
            return 0.0, Matching([])

        gram_matrix = self.inner.gram_matrix(x, y)
        problem = ILPMatchingProblem(x, y, gram_matrix, has_vars=True)
        problem.add_matching_constraint(self.constraint)
        problem.add_variable_matching_constraint()
        problem.add_latent_variable_constraint(self.cls)
        score, matches = problem.solve()
        return score, _matching_from_triples(original_x, original_y, score, x, y, matches)

    def score_self(self, x: Collection[T]) -> float:
        """Score a collection of objects with itself."""
        return SetMatchingMetric(self.inner, self.constraint).score_self(x)



"""`Path` object that has a string representation in JMESPath format.

This module provides a `Path` object that represents a path and corresponding methods to build and select paths.
The `Path` object has a string representation in JMESPath format. However, it does not support all JMESPath features.
"""

from dataclasses import dataclass
from typing import Union, SupportsInt, overload


def _index_to_int_str(i: int) -> str:
    return str(i) if i != -1 else "*"


def _int_str_to_index(s: str) -> int:
    return int(s) if s != "*" else -1


def _component_covers(selector: Union[str, int], component: Union[str, int]) -> bool:
    if isinstance(selector, int) and isinstance(component, int):
        return selector == component or selector == -1
    elif isinstance(selector, str) and isinstance(component, str):
        return selector == component or selector == "*"
    else:
        return False


@dataclass
class Path:
    """Represents a path selector. Its default string representation is in JMESPath format."""

    components: tuple[Union[str, int], ...] = ()  # [-1] means [*]

    @overload
    def __getitem__(self, item: SupportsInt) -> Union[str, int]: ...

    @overload
    def __getitem__(self, item: slice) -> tuple[Union[str, int], ...]: ...

    def __getitem__(self, item):
        """Returns a component of the path. This does not consider the root path."""
        if isinstance(item, slice):
            return self.components[item]
        elif isinstance(item, SupportsInt):
            return self.components[int(item)]
        else:
            raise NotImplementedError(f"Unsupported item type: {type(item)}")

    def is_root(self) -> bool:
        """Returns True if the path is the root path."""
        return len(self.components) == 0

    def __hash__(self):
        """Returns a hash of the path."""
        return hash(self.components)

    def __str__(self):
        """Returns a string representation of the path in JMESPath format."""
        if len(self.components) == 0:
            return "@"
        else:
            components = [
                f"[{_index_to_int_str(item)}]" if isinstance(item, int) else f".{item}" if i != 0 else item
                for i, item in enumerate(self.components)
            ]
            return "".join(components)

    def prepend(self, item: Union[str, int]) -> "Path":
        """Prepends an item to the path."""
        return Path((item,) + self.components)

    def append(self, item: Union[str, int]) -> "Path":
        """Appends an item to the path."""
        return Path(self.components + (item,))

    def selects(self, other: "Path") -> bool:
        """Returns True if the other path is selected by this path selector."""
        return len(self.components) == len(other.components) and all(
            _component_covers(sel, comp) for sel, comp in zip(self.components, other.components)
        )

    @classmethod
    def parse(cls, s: str):
        """Parses a string representation of the path in JMESPath format."""
        tokens = []
        t = ""
        for c in s:
            if c in ["@", ".", "[", "]"]:
                if t:
                    tokens.append(t)
                    t = ""
                tokens.append(c)
            else:
                t += c
        if t:
            tokens.append(t)
        components = []
        i = 0
        while i < len(tokens):
            if tokens[i] == "@":
                i += 1
                continue
            if tokens[i] == ".":
                i += 1
                continue
            elif tokens[i] == "[":
                components.append(_int_str_to_index(tokens[i + 1]))
                i += 3
            else:
                components.append(tokens[i])
                i += 1
        return Path(tuple(components))



from abc import ABC, abstractmethod
from dataclasses import dataclass, fields, is_dataclass
from typing import Any, Callable, Generic, Optional, TypeVar
from collections.abc import Collection, Iterator, Sequence

import numpy as np
import scipy as sp

from metametric.core.constraint import MatchingConstraint
from metametric.core._problem import MatchingProblem
from metametric.core.metric import Variable

T = TypeVar("T")


@dataclass
class ConstraintBuilder(ABC):
    n_x: int
    n_y: int
    n_x_vars: int
    n_y_vars: int

    def index_pair(self, i: int, j: int) -> int:
        return i * self.n_y + j

    def index_var_pair(self, i: int, j: int) -> int:
        return self.n_x * self.n_y + i * self.n_y_vars + j

    @abstractmethod
    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        pass


@dataclass
class MatchingConstraintBuilder(ConstraintBuilder):
    constraint: MatchingConstraint = MatchingConstraint.ONE_TO_ONE

    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        constraint_matrix_ctor: Callable[[int, int], Optional[np.ndarray]] = {
            MatchingConstraint.ONE_TO_ONE: _get_one_to_one_constraint_matrix,
            MatchingConstraint.ONE_TO_MANY: _get_one_to_many_constraint_matrix,
            MatchingConstraint.MANY_TO_ONE: _get_many_to_one_constraint_matrix,
            MatchingConstraint.MANY_TO_MANY: lambda _0, _1: None,
        }[self.constraint]
        m = constraint_matrix_ctor(self.n_x, self.n_y)
        if m is not None:
            m = np.concatenate(
                [
                    m,
                    np.zeros([m.shape[0], self.n_x_vars * self.n_y_vars]),
                ],  # pad with zeros for the latent variables
                axis=1,
            )
            return sp.optimize.LinearConstraint(
                A=m,
                ub=np.ones([m.shape[0]]),
            )
        else:
            return None  # no constraint when the constraint is MANY_TO_MANY


@dataclass
class VariableMatchingConstraintBuilder(ConstraintBuilder):
    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        if self.n_x_vars == 0 or self.n_y_vars == 0:
            return None
        # only one-to-one matching between variables
        m = _get_one_to_one_constraint_matrix(self.n_x_vars, self.n_y_vars)
        m = np.concatenate(
            [
                np.zeros([m.shape[0], self.n_x * self.n_y]),
                m,
            ],  # pad with zeros for the matching items
            axis=1,
        )
        return sp.optimize.LinearConstraint(
            A=m,
            ub=np.ones([m.shape[0]]),
        )


@dataclass
class MonotonicityConstraintBuilder(ConstraintBuilder):
    gram_matrix: np.ndarray  # R[n_x, n_y]
    x_reachability: np.ndarray  # R[n_x, n_x]
    y_reachability: np.ndarray  # R[n_y, n_y]

    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        vectors = []
        possible_matching_pairs = [
            (u, v) for u in range(self.n_x) for v in range(self.n_y) if self.gram_matrix[u, v] > 0
        ]
        for u0, v0 in possible_matching_pairs:
            for u1, v1 in possible_matching_pairs:
                if self.x_reachability[u0, u1] != self.y_reachability[v0, v1]:
                    vec = np.zeros(self.n_x * self.n_y)
                    vec[self.index_pair(u0, v0)] = 1
                    vec[self.index_pair(u1, v1)] = 1
                    vectors.append(vec)
                    # Enforce monotonicity of the matching
                    #    [u0 ~ v0] & [u1 ~ v1] -> [u0 <= u1] ≡ [v0 <= v1]
                    # => 1 - ((1 - t[u0~v0]) + (1 - t[u1~v1])) <= 1[[u0 <= u1] ≡ [v0 <= v1]]
                    # => t[u0~v0] + t[u1~v1] <= 1[[u0 <= u1] ≡ [v0 <= v1]] + 1
                    # => t[u0~v0] + t[u1~v1] <= 1  (if [u0 <= u1] ≡ [v0 <= v1], constraint redundant)

        if len(vectors) == 0:
            return None
        constraint_matrix = np.stack(vectors, axis=0)  # R[n_constraints, n_x * n_y]
        return sp.optimize.LinearConstraint(
            A=constraint_matrix,
            ub=np.ones([constraint_matrix.shape[0]]),
        )


@dataclass
class LatentVariableConstraintBuilder(ConstraintBuilder, Generic[T]):
    x: Collection[T]
    y: Collection[T]
    cls: type[T]
    gram_matrix: np.ndarray  # R[n_x, n_y]

    def __post_init__(self):
        assert is_dataclass(self.cls)

    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        x_vars = list(_all_variables(self.x))
        y_vars = list(_all_variables(self.y))
        x_var_to_id = {t: i for i, t in enumerate(x_vars)}
        y_var_to_id = {t: j for j, t in enumerate(y_vars)}
        vectors = []
        for i, a in enumerate(self.x):
            for j, b in enumerate(self.y):
                if self.gram_matrix[i, j] > 0:
                    for fld in fields(self.cls):  # pyright: ignore
                        a_fld = getattr(a, fld.name, None)
                        b_fld = getattr(b, fld.name, None)
                        if isinstance(a_fld, Variable) and isinstance(b_fld, Variable):
                            vec = np.zeros(self.n_x * self.n_y + self.n_x_vars * self.n_y_vars)
                            vec[self.index_pair(i, j)] = 1
                            vec[self.index_var_pair(x_var_to_id[a_fld], y_var_to_id[b_fld])] = -1
                            vectors.append(vec)
                            # Item matches implies variable matches
                            #    [a ~ b] -> [a_fld ~ b_fld]
                            # => t[a~b] <= t[a_fld~b_fld]
                            # => t[a~b] - t[a_fld~b_fld] <= 0
        if len(vectors) == 0:
            return None
        constraint_matrix = np.stack(vectors, axis=0)  # R[n_constraints, n_x * n_y + n_x_vars * n_y_vars]
        return sp.optimize.LinearConstraint(
            A=constraint_matrix,
            ub=np.zeros([constraint_matrix.shape[0]]),
        )


class ILPMatchingProblem(MatchingProblem[T]):
    """Creates a matching problem that is solved by ILP.

    The constrained ILP problem has variables
    - for each pair of elements in X and Y, and
    - for each pair of potential latent variables in X and Y.
    """

    def __init__(
        self,
        x: Sequence[T],
        y: Sequence[T],
        gram_matrix: np.ndarray,
        has_vars: bool = False,
    ):
        super().__init__(x, y, gram_matrix)
        self.n_x = len(x)
        self.n_y = len(y)
        if has_vars:
            self.x_vars = list(_all_variables(x))
            self.y_vars = list(_all_variables(y))
            self.n_x_vars = len(self.x_vars)
            self.n_y_vars = len(self.y_vars)
        else:
            self.x_vars = []
            self.y_vars = []
            self.n_x_vars = 0
            self.n_y_vars = 0
        self.constraints: list[sp.optimize.LinearConstraint] = []

    def add_matching_constraint(self, constraint_type: MatchingConstraint):
        constraint = MatchingConstraintBuilder(
            n_x=self.n_x,
            n_y=self.n_y,
            n_x_vars=self.n_x_vars,
            n_y_vars=self.n_y_vars,
            constraint=constraint_type,
        ).build()
        if constraint is not None:
            self.constraints.append(constraint)

    def add_variable_matching_constraint(self):
        constraint = VariableMatchingConstraintBuilder(
            n_x=self.n_x,
            n_y=self.n_y,
            n_x_vars=self.n_x_vars,
            n_y_vars=self.n_y_vars,
        ).build()
        if constraint is not None:
            self.constraints.append(constraint)

    def add_monotonicity_constraint(self, x_reachability: np.ndarray, y_reachability: np.ndarray):
        constraint = MonotonicityConstraintBuilder(
            n_x=self.n_x,
            n_y=self.n_y,
            n_x_vars=self.n_x_vars,
            n_y_vars=self.n_y_vars,
            gram_matrix=self.gram_matrix,
            x_reachability=x_reachability,
            y_reachability=y_reachability,
        ).build()
        if constraint is not None:
            self.constraints.append(constraint)

    def add_latent_variable_constraint(self, cls: type[T]):
        constraint = LatentVariableConstraintBuilder(
            n_x=self.n_x,
            n_y=self.n_y,
            n_x_vars=self.n_x_vars,
            n_y_vars=self.n_y_vars,
            x=self.x,
            y=self.y,
            cls=cls,
            gram_matrix=self.gram_matrix,
        ).build()
        if constraint is not None:
            self.constraints.append(constraint)

    def solve(self):
        # Layout of the constraint matrix:
        # ╒═════════════════════════╤═══════════════════════════╕
        # │        n_x * n_y        │    n_x_vars * n_y_vars    │
        # ┝━━━━━━━━━━━━━━━━━━━━━━━━━┿━━━━━━━━━━━━━━━━━━━━━━━━━━━┥
        # │   MATCHING_CONSTRAINT   │                           │
        # ├─────────────────────────┼───────────────────────────┤
        # │                         │  VAR_MATCHING_CONSTRAINT  │
        # ├─────────────────────────┼───────────────────────────┤
        # │ MONOTONICITY_CONSTRAINT │                           │
        # ├─────────────────────────┴───────────────────────────┤
        # │            LATENT_VARIABLE_CONSTRAINT               │
        # ╘═════════════════════════════════════════════════════┙
        coef = np.concatenate(
            [
                self.gram_matrix.reshape([self.n_x * self.n_y]),
                np.zeros([self.n_x_vars * self.n_y_vars]),
            ],  # pad with zeros for the latent variables
            axis=0,
        )
        result = sp.optimize.milp(
            c=-coef,
            constraints=self.constraints,
            bounds=sp.optimize.Bounds(lb=0, ub=1),
            integrality=np.ones_like(coef),
        )
        solution = result.x[: self.n_x * self.n_y].reshape([self.n_x, self.n_y])
        matching = [
            (i, j, self.gram_matrix[i, j].item())
            for i in range(self.n_x)
            for j in range(self.n_y)
            if solution[i, j] > 0
        ]
        return -result.fun, matching


def _get_one_to_many_constraint_matrix(n_x: int, n_y: int) -> np.ndarray:  # [Y, X * Y]
    mask_y = np.zeros([n_y, n_x, n_y])
    mask_y[np.arange(n_y), :, np.arange(n_y)] = 1
    mask_y = mask_y.reshape([n_y, n_x * n_y])
    return mask_y


def _get_many_to_one_constraint_matrix(n_x: int, n_y: int) -> np.ndarray:  # [X, X * Y]
    mask_x = np.zeros([n_x, n_x, n_y])
    mask_x[np.arange(n_x), np.arange(n_x), :] = 1
    mask_x = mask_x.reshape([n_x, n_x * n_y])
    return mask_x


def _get_one_to_one_constraint_matrix(n_x: int, n_y: int) -> np.ndarray:  # [X + Y, X * Y]
    mask_x = _get_one_to_many_constraint_matrix(n_x, n_y)
    mask_y = _get_many_to_one_constraint_matrix(n_x, n_y)
    return np.concatenate([mask_x, mask_y], axis=0)


def _all_variables(obj: Any) -> Iterator[Variable]:
    if isinstance(obj, Variable):
        yield obj
    elif isinstance(obj, Collection) and not isinstance(obj, str):
        for item in obj:
            yield from _all_variables(item)
    elif getattr(obj, "__dict__", None) is not None:
        for fld in vars(obj).values():
            yield from _all_variables(fld)



"""The core functionality of metametric package."""

__version__ = "0.2.0rc0"

from metametric.core.metric import Metric, Variable  # noqa: F401
from metametric.core.reduction import Reduction  # noqa: F401
from metametric.core.metric_suite import MetricSuite, MetricFamily  # noqa: F401
from metametric.core.constraint import MatchingConstraint  # noqa: F401
from metametric.core.path import Path  # noqa: F401



"""Defines utilities for obtaining the inner matching after a metric is computed."""

from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Callable, Any
from collections.abc import Iterable
from dataclasses import dataclass

from metametric.core.path import Path

T = TypeVar("T", covariant=True)
Tc = TypeVar("Tc", contravariant=True)


@dataclass
class Match(Generic[T]):
    """Represents a match between a pair of inner objects in a prediction and a reference."""

    pred_path: Path
    pred: T
    ref_path: Path
    ref: T
    score: float

    def __str__(self):
        """Returns a string representation of the match."""
        return f"{self.pred_path} -> {self.ref_path} ({self.score})"


class Hook(ABC, Generic[Tc]):
    """A hook that is called when a match is found."""

    @abstractmethod
    def on_match(self, data_id: int, pred_path: str, pred: Tc, ref_path: str, ref: Tc, score: float):
        """Called when a match is found."""
        raise NotImplementedError

    @staticmethod
    def from_callable(func: Callable[[int, str, Tc, str, Tc, float], None]) -> "Hook[Tc]":
        """Creates a hook from a callback."""
        return _HookFromCallable(func)


class _HookFromCallable(Hook[Tc]):
    def __init__(self, func: Callable[[int, str, Tc, str, Tc, float], None]):
        self.func = func

    def on_match(self, data_id: int, pred_path: str, pred: Tc, ref_path: str, ref: Tc, score: float):
        self.func(data_id, pred_path, pred, ref_path, ref, score)


class Matching(Iterable[Match[object]]):
    """An object that can be used to iterate over matches and run hooks on them."""

    def __init__(self, matches: Iterable[Match[object]]):
        self.matches = matches

    def __iter__(self):
        """Traverses all matching pairs of inner objects."""
        return iter(self.matches)

    def run_with_hooks(self, hooks: dict[str, Hook[Any]], data_id: int = 0):
        """Runs hooks on the matches."""
        parsed_hooks = {Path.parse(selector): hook for selector, hook in hooks.items()}
        for match in self.matches:
            for selector, hook in parsed_hooks.items():
                if selector.selects(match.pred_path):
                    hook.on_match(
                        data_id, str(match.pred_path), match.pred, str(match.ref_path), match.ref, match.score
                    )



"""Contains metrics for event extraction."""

import metametric.dsl as mm

from metametric.structures.ie import Event, EventSet, Trigger, Argument

trigger_identification = mm.dataclass[EventSet](
    {
        "events": mm.set_matching[Event, "<->"](
            mm.dataclass[Event](
                {
                    "trigger": mm.dataclass[Trigger](
                        {
                            "mention": ...,
                        }
                    ),
                }
            )
        )
    }
)

trigger_classification = mm.dataclass[EventSet](
    {
        "events": mm.set_matching[Event, "<->"](
            mm.dataclass[Event](
                {
                    "trigger": ...,
                }
            )
        )
    }
)

argument_identification = mm.dataclass[EventSet](
    {
        "events": mm.set_matching[Event, "<->"](
            mm.dataclass[Event](
                {
                    "trigger": ...,
                    "args": mm.set_matching[Argument, "<->"](
                        mm.dataclass[Argument](
                            {
                                "mention": ...,
                            }
                        )
                    ),
                }
            )
        )
    }
)

argument_classification = mm.dataclass[EventSet](
    {
        "events": mm.set_matching[Event, "<->"](
            mm.dataclass[Event]({"trigger": ..., "args": mm.set_matching[Argument, "<->"](...)})
        )
    }
)


event_extraction_suite = mm.suite(
    {
        "trigger_identification": mm.family(trigger_identification, mm.macro_average(["precision", "recall", "f1"])),
        "trigger_classification": mm.family(trigger_classification, mm.macro_average(["precision", "recall", "f1"])),
        "argument_identification": mm.family(argument_identification, mm.macro_average(["precision", "recall", "f1"])),
        "argument_classification": mm.family(argument_classification, mm.macro_average(["precision", "recall", "f1"])),
    }
)



"""Metric interface and implementations for commonly used metrics."""

from abc import abstractmethod
from dataclasses import dataclass, is_dataclass
from functools import reduce
from operator import mul
from typing import (
    Callable,
    ClassVar,
    Generic,
    Protocol,
    TypeVar,
    Union,
    get_origin,
    runtime_checkable,
)
from collections.abc import Sequence

import numpy as np

from metametric.core.matching import Matching, Match, Path

S = TypeVar("S", contravariant=True)
T = TypeVar("T", contravariant=True)
U = TypeVar("U")


class Metric(Generic[T]):
    r"""The basic metric interface.

    Here a *metric* is defined as a function $\phi: T \times T \to \mathbb{R}_{\ge 0}$ that takes two objects and
    returns a non-negative number that quantifies their similarity.
    It follows the common usage in machine learning and NLP literature, as in the phrase "evaluation metrics".
    This is *not* the metric in the mathematical sense, where it is a generalization of *distances*.
    """

    @abstractmethod
    def compute(self, x: T, y: T) -> tuple[float, Matching]:
        r"""Scores two objects using this metric, and returns the score and a matching object."""
        raise NotImplementedError

    def score(self, x: T, y: T) -> float:
        r"""Scores two objects using this metric: $\phi(x, y)$."""
        return self.compute(x, y)[0]

    def score_self(self, x: T) -> float:
        r"""Scores an object against itself: $\phi(x, x)$.

        In many cases there is a faster way to compute this than the general pair case.
        In such cases, please override this function.
        """
        return self.score(x, x)

    def gram_matrix(self, xs: Sequence[T], ys: Sequence[T]) -> np.ndarray:
        r"""Computes the Gram matrix of the metric given two collections of objects.

        Args:
            xs: A collection of objects $\{x_1, \ldots, x_n\}$.
            ys: A collection of objects $\{y_1, \ldots, y_m\}$.

        Returns:
            A Gram matrix $G$ where $G = \begin{bmatrix} \phi(x_1, y_1) & \cdots & \phi(x_1, y_m) \\
                                \vdots & \ddots & \vdots \\
                                \phi(x_n, y_1) & \cdots & \phi(x_n, y_m) \end{bmatrix}$.
        """
        return np.array([[self.score(x, y) for y in ys] for x in xs])

    def contramap(self, f: Callable[[S], T]) -> "Metric[S]":
        r"""Returns a new metric $\phi^\prime$ by first preprocessing the objects by a given function $f: S \to T$.

        \[ \phi^\prime(x, y) = \phi(f(x), f(y)) \]

        Args:
            f: A preprocessing function.

        Returns:
            A new metric $\phi^\prime$.

        """
        return ContramappedMetric(self, f)

    @staticmethod
    def from_function(f: Callable[[T, T], float]) -> "Metric[T]":
        r"""Create a metric from a function $f: T \times T \to \mathbb{R}_{\ge 0}$.

        Args:
            f (`Callable[[T, T], float]`):
                A function that takes two objects and returns a float.
                This is the function that derives the metric.

        Returns:
            `Metric[T]`: A metric that uses the function to score two objects.
        """
        return MetricFromFunction(f)


class MetricFromFunction(Metric[T]):
    """A metric wrapped from a function."""

    def __init__(self, f: Callable[[T, T], float]):
        self.f = f

    def compute(self, x: T, y: T) -> tuple[float, Matching]:
        """Score two objects."""
        score = self.f(x, y)

        def _matching():
            yield Match(Path(), x, Path(), y, score)

        return self.f(x, y), Matching(_matching())


class ContramappedMetric(Metric[S]):
    """A metric contramapped by a function."""

    def __init__(self, inner: Metric[T], f: Callable[[S], T]):
        self.inner = inner
        self.f = f

    def compute(self, x: S, y: S) -> tuple[float, Matching]:
        """Score two objects."""
        return self.inner.compute(self.f(x), self.f(y))

    def score_self(self, x: S) -> float:
        """Scores an object against itself."""
        return self.inner.score_self(self.f(x))


class DiscreteMetric(Metric[T]):
    """A metric for discrete objects."""

    def __init__(self, cls: type[T]):
        if getattr(cls, "__eq__", None) is None:
            raise ValueError("Class must implement __eq__")

    def compute(self, x: T, y: T) -> tuple[float, Matching]:
        """Score two objects."""
        if x == y:
            return 1.0, Matching([Match(Path(), x, Path(), y, 1.0)])
        return 0.0, Matching([])

    def score_self(self, x: T) -> float:
        """Scores an object against itself."""
        return 1.0


class ProductMetric(Metric[T]):
    """A metric that is the product of other metrics."""

    def __init__(self, cls: type[T], field_metrics: dict[str, Metric]):
        if not is_dataclass(cls):
            raise ValueError(f"{cls} has to be a dataclass.")
        self.field_metrics = field_metrics

    def compute(self, x: T, y: T) -> tuple[float, Matching]:
        """Score two objects."""
        field_scores = {
            fld: self.field_metrics[fld].compute(getattr(x, fld), getattr(y, fld)) for fld in self.field_metrics.keys()
        }
        total_score = reduce(mul, (s for s, _ in field_scores.values()), 1.0)

        def _matching():
            yield Match(Path(), x, Path(), y, total_score)
            for fld, (s, matching) in field_scores.items():
                for m in matching.matches:
                    yield Match(m.pred_path.prepend(fld), m.pred, m.ref_path.prepend(fld), m.ref, m.score)

        return total_score, Matching(_matching())


class UnionMetric(Metric[T]):
    """A metric that is the union of other metrics."""

    def __init__(self, cls: type[T], case_metrics: dict[type, Metric]):
        if get_origin(cls) is not Union:
            raise ValueError(f"{cls} has to be a union.")
        self.case_metrics = case_metrics

    def compute(self, x: T, y: T) -> tuple[float, Matching]:
        """Score two objects."""
        x_type = type(x)
        y_type = type(y)
        if x_type != y_type:
            return 0.0, Matching([])
        return self.case_metrics[x_type].compute(x, y)

    def score_self(self, x: T) -> float:
        """Scores an object against itself."""
        return 1.0


@dataclass(eq=True, frozen=True)
class Variable:
    """A variable in latent matchings."""

    name: str

    latent_metric: ClassVar[Metric["Variable"]] = Metric.from_function(lambda x, y: 1.0)


@runtime_checkable
class HasMetric(Protocol[T]):
    """Protocol for classes that have a metric."""

    metric: Metric[T]


@runtime_checkable
class HasLatentMetric(Protocol[T]):
    """Protocol for classes that have a latent metric."""

    latent_metric: Metric[T]
